"""
#######################################################################
#
# Licensed Materials - Property of Pipelinefx L.L.C.
#
#
# (C) COPYRIGHT Pipelinefx Limited Liability Corporation.
#  All Rights Reserved.
#                               
#  US Government Users Restricted Rights - Use, duplication or
#  disclosure restricted by GSA ADP Schedule Contract with 
#  Pipelinefx L.L.C.
#
# QubeView Frame module
#
# Name: qbUIFrame.py
#
# $DateTime: 2012/06/12 15:12:47 $
# $Revision: #1 $
# $Change: 9701 $
# $Id: //depot/main/qubeMobileServer/qubeMobile/model/qbUIFrame.py#1 $
# $File: //depot/main/qubeMobileServer/qubeMobile/model/qbUIFrame.py $
#
# Written by: Brian Knudson / PFX
#             briank@pipelinefx.com
#             2011-10-01
#
# Description:
#  Frame module, subclassed off of qb.Work with some extra functions to
#  access logs using local cache
#
#
#######################################################################
"""

import os
import sys
import re
import time
import datetime
import mmap
import logging
import tempfile

try:
    import qb
except ImportError:
    api_path = "/Applications/pfx/qube/api/python/"
    if api_path not in sys.path:
        sys.path.insert(0,api_path)
    import qb


class QbUIFrame(qb.Work):
    """
    A frame model.
    Construct with:
     frame: a qb.Work object
     log_list: a qbUILogList
    Note: Both outfp & errfp can be None, but they will never be updated in that case
    """
    def __init__(self,frame,log_list):
        super(QbUIFrame,self).__init__(frame)
        self.__log_list = log_list


    def getLogPaths(self):
        """
        Returns the paths to the temporary files used for this frame's logs
        return is (stdout_path,stderr_path)
        """
        return self.__log_list[self['pid']].getSubjobLogPath(self['subid'])


    def getLogs(self,update=False):
        """
        Return the log for this one frame
        return val is (stdout,stderr)
        """
        logging.debug("Fetching logs for frame '%s:%s'." % (self['pid'],self['id']))
        return self.__log_list[self['pid']].getFrameLogs(self['id']-1) # -1 to make up for 1 based array in the qb API


    def getReadableElapsed(self):
        """
        returns the time this frame has been running in a human
        readable string.
        """
        now = int(time.time())
        start_epoch    = {True:self['timestart'],    False:now}[self['timestart'] > qb.QB_TIME_EPOCH_OFFSET ]
        complete_epoch = {True:self['timecomplete'], False:now}[self['timecomplete'] > qb.QB_TIME_EPOCH_OFFSET ]
        elapsed_epoch = complete_epoch - start_epoch

        return {True:str(datetime.timedelta(seconds=elapsed_epoch)), False:"--" }[elapsed_epoch > 0]


    def getAttribute(self,att):
        """
        returns the attribute from this instance, given an attribute or human
        readable name.
        """
        att = att.lower()
        if self.has_key(att):
            return self[att]
        else:
            if att == 'started':
                return { True:time.strftime("%X",time.localtime(self['timestart'])), False:"--" }[ self['timestart'] > qb.QB_TIME_EPOCH_OFFSET ]
            elif att == 'completed':
                return { True:time.strftime("%X",time.localtime(self['timecomplete'])), False:"--" }[ self['timecomplete'] > qb.QB_TIME_EPOCH_OFFSET ]
            elif att == 'elapsed':
                return self.getReadableElapsed()
            else:
                return att
