"""
#######################################################################
#
# Licensed Materials - Property of Pipelinefx L.L.C.
#
#
# (C) COPYRIGHT Pipelinefx Limited Liability Corporation.
#  All Rights Reserved.
#                               
#  US Government Users Restricted Rights - Use, duplication or
#  disclosure restricted by GSA ADP Schedule Contract with 
#  Pipelinefx L.L.C.
#
# $DateTime: 2012/10/12 16:20:55 $
# $Revision: #5 $
# $Change: 10156 $
# $Id: //depot/main/qubeMobileServer/qubeMobile/model/qbUIJob.py#5 $
# $File: //depot/main/qubeMobileServer/qubeMobile/model/qbUIJob.py $
#
# Written by: Brian Knudson / PFX
#             briank@pipelinefx.com
#             2011-10-05
#
# Description:
#  UIJob class - extends qb.job & provides a few more convenience methods
#  suited for UIs
#
#
#######################################################################
"""

import os
import sys
import re
import time
import glob
import datetime
import tempfile
import logging

try:
    import qb
except ImportError:
    api_path = "/Applications/pfx/qube/api/python/"
    if api_path not in sys.path:
        sys.path.insert(0,api_path)
    import qb

from qbUIFrame import QbUIFrame
from qbUISubjob import QbUISubjob


# class QbUIJob(qb.Job):
class QbUIJob(dict):
    """
    QbUIJob is a subclass of qb.Job.  It defines additional properties and methods
    that make use with UIs more convenient.

    Poperties:

     - frames:  Returns a list of agenda items for this job.  If the list doesn't
                yet exist, it is created then returned.  Each agenda item is a
                qb.Work object.
     - subjobs: Returns a list of subjobs from this job. If the list doesn't yet
                exist, it is created then returned.  Each subjob is a qb.Subjob
                object.

    Methods:

     - updateDetail:          Fetches the agenda and subjobs list, and callbacks.
                              Updates a timestamp noting when this happened.
     - hasDetail:             Returns true if this job has updated agenda & subjob
                              lists.
     - getLogs:               Returns the logs for this job. See method docstring
                              for details
     - getAttribute:          Returns the value for a given attribute
     - calculateAvgFrameTime: Returns the average time to complete an agenda item
     - calculateETA:          Returns a datetime object of the estimated time of
                              completion
     - getHumanReadableTime:  Returns the human readable version of any of this
                              job's time stamps.

    --== qb.Job docstring follows ==--

    """
    try:
        __doc__ += qb.Job.__doc__
    except:
        pass
    def __init__(self,job,log_list=None,query_SQL=False):
        super(QbUIJob,self).__init__(job)
        self.last_ui_update = int(time.time())

        # if we were called with one of our own, this should work:
        if not log_list and hasattr(job,'getLogList'):
            self.__log_list = job.getLogList()
        else:
            self.__log_list = log_list

        self.__frames = []
        self.__subjobs = []
        self.__has_detail = False

        if query_SQL:
            import qb.query
            self.qb_jobinfo = qb.query.jobinfo
        else:
            import qb # For windows
            self.qb_jobinfo = qb.jobinfo



    @property
    def frames(self):
        """
        If the list of frames doesn't exist, it populates the list.
        Otherwise, it returns the list
        """
        if not self.__frames:
            self.__updateFrames()
        return self.__frames


    @property
    def subjobs(self):
        """
        If the list of subjobs doesn't exist, it populates the list.
        Otherwise, it returns the list
        """
        if not self.__subjobs:
            self.__updateSubjobs()
        return self.__subjobs


    def __updateFrames(self):
        """
        Fills the internal frames list.  This is very similar to the
        'agenda' list, though it will have some extra UI stuff
        """
        logging.info("updating frames for job '%s'" % self['id'])
        if not self.__has_detail:
            self.updateDetail(frames=False,subjobs=False)
        self.__frames = []
        for a in self.get('agenda',[]):
            self.__frames.append(QbUIFrame(a,log_list=self.__log_list))


    def __updateSubjobs(self):
        """
        Fills the internal subjobs list.  This is very similar to the
        'agenda' list, though it will have some extra UI stuff
        """
        logging.info("updating subjobs for job '%s'" % self['id'])
        if not self.__has_detail:
            self.updateDetail(subjobs=False)
        self.__subjobs = []
        for sj in self.get('subjobs',[]):
            self.__subjobs.append(QbUISubjob(subjob=sj,log_list=self.__log_list))



    def __getUnfavorableCount(self):
        """
        Return the number of work items that are in a negative state (failed, blocked, badlogin, etc)
        """
        ret = 0
        ignore_fields = ('complete','running','pending')
        return sum((v for k,v in self['todotally'].items() if k not in ignore_fields))


    def __getFavorableCount(self):
        """
        Return the number of work items that are in a positive/potential state (complete, running, pending)
        """
        ret = 0
        use_fields = ('complete','running','pending')
        return sum((v for k,v in self['todotally'].items() if k in use_fields))

    
    def __getWorkCount(self):
        """
        Return the total number of work items for this job
        """
        return sum(self['todotally'].values())


    def updateDetail(self,frames=True,subjobs=True):
        """
        Updates the current job, taking data from the supervisor.
        The updated job will now have agenda, subjob, and callback info.
        A user will likely never have to call this function, though it is exposed in then
        event you need to.  Requesting a frame from this job will automatically call
        this function.
        If 'frames' is false, this function will not update the frame list (False is the default)
        Returns itself for faster access to data, but the return value is not
        strictly necessary as the object is now updated & all of the data in
        the return is also available in the object
        """
        logging.info("updating job detail for job '%s'" % self['id'])
        self.last_ui_update = int(time.time())
        try:
            super(QbUIJob,self).__init__(self.qb_jobinfo(id=self['id'],fields=['agenda', 'subjobs', 'callbacks'])[0])
        except IndexError:
            logging.error("Job %d no longer appears to exist." % self['id'])
            return None # might still want to return self

        self.__has_detail = True

        if frames:
            self.__updateFrames()
        if subjobs:
            self.__updateSubjobs()

        return self
    

    def hasDetail(self):
        """
        Accessor function for the private 'has_detail'
        returns the val of the var
        """
        return self.__has_detail


    def getLogs(self,*subjobs):
        """
        Get logs for this job based on give subjob(s).  Subjobs should be given as
        a comma separated list.
        If no subjob is given, logs for all subjobs are returned.
        Return value is a tuple of lists, return[0] is stdout, return[1] is stderr.
        Each item in the stdout/err list is a subjob log dictionary, as returned
        by qb.stdout/err(), which has the form of {jobid:'',subid:'',data''} where
        'data' is the actual log.  This behaves much like qb.stdout,qb.stderr.
        """
        logging.info("Getting logs for job '%s'" % self['id'])
        return self.__log_list[self['id']].getJobLogs()


    def getAttribute(self,att):
        """
        returns the attribute from this instance, given an attribute or human
        readable name.
        """
        att = att.lower()
        if self.has_key(att):
            return self[att]
        else:
            if att == 'jobid':
                return self['id']
            elif att == 'complete':
                return self['todotally']['complete']
            elif att == 'running':
                return self['todotally']['running']
            elif att == 'pending':
                return self['todotally']['pending']
            elif att == 'failed':
                return self['todotally']['failed']
            elif att == 'badlogin':
                return self['todotally']['badlogin']
            elif att == 'blocked':
                return self['todotally']['blocked']
            elif att == 'suspended':
                return self['todotally']['suspended']
            elif att == 'unknown':
                return self['todotally']['unknown']
            elif att == 'waiting':
                return self['todotally']['waiting']
            elif att == 'range':
                frange = self.get('package',{}).get('range')
                if not frange:
                    frange = self.get('package',{}).get('frameCount')
                    if frange: frange = "1-"+frange
                return frange
            elif att == 'submitted':
                return self.getHumanReadableTime('submit')
            elif att == 'started':
                return self.getHumanReadableTime('start')
            elif att == 'completed':
                return self.getHumanReadableTime('complete')
            elif att == 'avg time' or att == 'average frame':
                return self.calculateAvgFrameTime()
            elif att == 'job eta':
                eta = self.calculateETA()
                if eta: eta = self.getHumanReadableTime(time_stamp=int(eta.strftime("%s")))
                return eta
            elif att == 'progress':
                # only return *text* progress if we have frames and we're not running
                if sum(self['todotally'].values())>0:
                    return "%d%%" % (int( (float (self['todotally']['complete'])/float(sum(self['todotally'].values()))) * 100 ))
                elif sum(self['cpustally'].values())>0:
                    return "%d%%" % (int( (float (self['cpustally']['complete'])/float(sum(self['cpustally'].values()))) * 100 ))
                else:
                    return "0%"
            elif att == 'slots':
                return "%s/%s" % (self['todotally']['running'],self['cpus'])
            else:
                return att
            
        

    def calculateAvgFrameTime(self):
        """
        Calculate Average Frame Times from the Tasks or Subjobs
        """
        # Note: This function should not be used in the joblist.  if it is... it's going to get a detail
        # for EVERY job in the queue.... BAD!  Maybe there should be some logic to prevent this
        if not self.__has_detail:
            self.updateDetail()

        # Use either Tasks or Subjobs if no tasks
        if self['todo'] > 0:   # Use Tasks
            itemField = 'agenda'
            tallyField = 'todotally'
        else: # Use Subjobs: No agenda items, so must just have subjobs
            itemField = 'subjobs'
            tallyField = 'cpustally'

        try:
            agenda = self[itemField]
        except KeyError:
            return None
        elapsedSeconds_avg = 0
        numCompleteTasks = self[tallyField]['complete']
        if numCompleteTasks > 0:
            for a in agenda:
                if a['status'] == 'complete':
                    elapsedSeconds = a['timecomplete'] - a['timestart']
                    elapsedSeconds_avg += elapsedSeconds
            elapsedSeconds_avg /= numCompleteTasks
            return datetime.timedelta(seconds=elapsedSeconds_avg)
        else:
            return None


    def calculateETA(self, now=None):
        """
        determine ETA for this job
        """
        if not self.__has_detail:
            self.updateDetail()

        # if not running, no need to calc an ETA
        if self['status'] != 'running':
            return None

        # if we can't figure an average frame time, no need to calc ETA
        avgFrame = self.calculateAvgFrameTime()
        if not avgFrame:
            return None

        # Use either Tasks or Subjobs if no tasks
        if self['todo'] > 0:   # Use Tasks
            itemField = 'agenda'
            tallyField = self['todotally']
        else: # Use Subjobs: No agenda items, so must just have subjobs
            itemField = 'subjobs'
            tallyField = self['cpustally']
    
        if tallyField['complete'] > 0 and tallyField['running'] > 0: # has both complete items and is running
            if not now:
                now = datetime.datetime.now()
            # Multiply pending by frame average
            eta  = avgFrame * tallyField['pending']
            eta += avgFrame * tallyField['running']
            # Reduce by running frames elapsed time (cap at avg frame time)
            agenda = self.get(itemField, [])
            for a in agenda:
                if a['status'] == 'running':
                    runningDuration = now - datetime.datetime.fromtimestamp(a['timestart'])
                    eta -= min(runningDuration,avgFrame)
            eta /= tallyField['running']
            now -= datetime.timedelta(microseconds=now.microsecond)
            eta -= datetime.timedelta(microseconds=eta.microseconds)
            return now+eta
        else:
            return None # datetime.datetime.fromtimestamp(0)  #datetime.timedelta(seconds=0) #'' #'inactive' (not running or pending)
    
    def getHumanReadableTime(self,time_type='start',time_stamp=None):
        """
        Returns a string representation of one of the job timestamps.
        time_type can be 'submit', 'start', or 'complete'. Default is 'start'
        if time_stamp is given, time_type is ignored
        """
        time_key = "time%s" % time_type
        epoch = self[time_key]
        if time_stamp:
            epoch = time_stamp

        # qb.QB_TIME_EPOCH_OFFSET is the default value for something that hasn't yet fired.
        if epoch == qb.QB_TIME_EPOCH_OFFSET: return None

        # If we're in the same day, just restart the time of the event, otherwise, include the date
        if time.localtime()[:3] == time.localtime(epoch)[:3]:
            # could use %X here, but, being compiled, might always be the US representation
            return time.strftime("%H:%M:%S",time.localtime(epoch))  
        return time.strftime("%Y-%m-%d %H:%M:%S",time.localtime(epoch))
            

    def kill(self):
        """
        Kill this job
        """
        return qb.kill(self['id'])

    def retry(self,only_failed=False):
        """
        Retry all frames in this job
        """
        if only_failed:
            frames = ['%d:%d' % (f['pid'],f['id']) for f in self.frames if f['status'] == 'failed']
        else:
            frames = ['%d:%d' % (f['pid'],f['id']) for f in self.frames]

        if frames:
            qb.retrywork(*frames)
            return qb.retry(self['id'])
        else:
            return None
        
