"""
#######################################################################
#
# Licensed Materials - Property of Pipelinefx L.L.C.
#
#
# (C) COPYRIGHT Pipelinefx Limited Liability Corporation.
#  All Rights Reserved.
#                               
#  US Government Users Restricted Rights - Use, duplication or
#  disclosure restricted by GSA ADP Schedule Contract with 
#  Pipelinefx L.L.C.
#
# $DateTime: 2012/10/12 16:20:55 $
# $Revision: #4 $
# $Change: 10156 $
# $Id: //depot/main/qubeMobileServer/qubeMobile/model/qbUIJobList.py#4 $
# $File: //depot/main/qubeMobileServer/qubeMobile/model/qbUIJobList.py $
#
# Written by: Brian Knudson / PFX
#             briank@pipelinefx.com
#             2011-10-05
#
# Description:
#  UIJobList class - Holds a list of a qbUIJobs with sensible access
#  to job methods.
#
#
#######################################################################
"""

import os
import sys
import re
import time
import datetime
import logging
from operator import itemgetter

try:
    import qb
except ImportError:
    for api_path in ("/Applications/pfx/qube/api/python/",
                     "/usr/local/pfx/qube/api/python/",
                     "C:\\Program Files (x86)\\pfx\\qube\\api\\python"):
        if api_path not in sys.path and os.path.exists(api_path):
            sys.path.insert(0,api_path)
            break
    import qb


from qbUIList import QbUIList
from qbUIJob import QbUIJob
from qbUILogList import QbUILogList

"""
Qb UI JobList Class
"""
    


class QbUIJobList(QbUIList):
    """
    A job list model.
    Can be contructed:
      blank: which will get all hosts from the db
      list: will create an internal dict from the list of hosts
      dict: will copy the given dict to internal
    """
    def __init__(self, filters={}, job_list=[], limit=50, verbosity=0):
        logging.debug("Creating QbUIJobList")
        super(QbUIJobList,self).__init__(job_list,verbosity)

        self.limit=limit

        self.query_SQL = True
        try:
            import qb.query
            db = qb.query.mysqlConnect()
            dbHostInfo = db.get_host_info()
            db.close()
            self.qb_jobinfo = qb.query.jobinfo
        except:
            import qb # for Windows
            self.qb_jobinfo = qb.jobinfo
            self.query_SQL = False

        # if a cient_logpath is given, pass it to the jobs (doing it here means
        # one stat for all jobs rather than one stat per job if it was done in the
        # job...  being in the job seems to make more OO sense, but.. too expensive)
        try:
            self.__mounted_logs = os.path.exists(qb.localconfig().get("client_logpath"))
            if self.__mounted_logs:
                logging.info("Use logfile location specified by client_ligpath: '%s'" % qb.localconfig().get("client_logpath"))
            else:
                logging.info("Fetching logfiles from supervisor")
        except:
            logging.info("Fetching logfiles from supervisor")
            self.__mounted_logs = False


        self.__log_list = QbUILogList(job_list=self,query_SQL=self.query_SQL)

        # NOTE: This is weird logic... designed to fit the class = eval(repr(class)) idea.
        #  Probably needs a re-think.
        # if we were passed a dict, we've already ini'd our parent, so we're done,
        #  otherwise, we've got some updatin' to do:
        #if not isinstance(job_list,dict):
        if not self._data:
            self.updateJobList(filters=filters,job_list=job_list,clear=True)



    ###
    #
    #  Convenience methods (not coming from base class):
    #
    def __str__(self):
        """
        str(object) -> string

        Return a nice string representation of the object.
        If the argument is a string, the return value is the same object.
        """
        keys = sorted(self._data.keys())
        start = end = 0
        if keys:
            start = keys[0]
            end = keys[-1]
        return "QbUIJobList. size: %d, id rage: %d-%d, last updated: %s" % (len(self._data),
                                                                            start,
                                                                            end,
                                                                            time.ctime(self.last_updated))

    def __del__(self):
        """
        Destructor that makes sure we explcitly clean up
        """
        logging.info("Joblist destructor...")
        self.clearAll()

    ###
    #
    #  Custom methods for this class:
    #
    def clearAll(self):
        """
        Explicitly calls the clearCache method of all jobs.  It appears that their
        destructor is not always called, therefore leaving cache files in tmp.  This
        prevents that.
        Also explicitly clears the job (_data) dict.
        """
        logging.info("Job list clearing all log cache files...")
        self.__log_list.clearAll()
        #self._data.clear() # maybe redundant?
        self._data = {}
        self.last_updated = 0


    def getLogList(self):
        """
        Returns the loglist model.  Mainly used for debugging
        """
        return self.__log_list
    

    def updateJobList(self,filters={},clear=False,job_list=None):
        """
        Populates internal dictionary with the return of an API call.
        Params:
          clear=False:  If clear is True, the internal dict is cleared before being populated
                        If clear is False (default), only the entries in the internal dict that
                         are newer than our internal timestamp are updated.
          jobList=None: If supplied, use this last rather than fetching one via the api.
        Pre-condition: None
        Post-condition: internal dictionary has complete of partial new data
        Return value: None
        """
        start = time.time()
        if clear:
            self.clearAll()

        # get the time before the call, in the event the call takes forever to return
        now = int(time.time())

        if job_list:
            # if we were given a list of jobs, check to see if we need to
            # update our last_updated
            self.last_updated = max(j.get('lastupdate',0) for j in job_list.values())
        else:
            if self.query_SQL:
                job_list = self.qb_jobinfo(filters=filters,updatedAfter=self.last_updated,limit=self.limit,sortOrder='DESC')
            else:
                job_list = sorted(self.qb_jobinfo(filters=filters),key=itemgetter('id'),reverse=True)[:self.limit]
            self.last_updated = now

        # updated the list with what we just got.  Dirty up the entry so we know it's new
        for j in job_list:
            self._data[j['id']] = QbUIJob(j,log_list=self.__log_list,query_SQL=self.query_SQL)

        end = time.time()

        logging.info("Updated %d job(s) in %.2fs" % (len(job_list),end-start))

        

