
"""
#######################################################################
#
# Licensed Materials - Property of Pipelinefx L.L.C.
#
#
# (C) COPYRIGHT Pipelinefx Limited Liability Corporation.
#  All Rights Reserved.
#                               
#  US Government Users Restricted Rights - Use, duplication or
#  disclosure restricted by GSA ADP Schedule Contract with 
#  Pipelinefx L.L.C.
#
# $DateTime: 2012/10/16 15:23:27 $
# $Revision: #5 $
# $Change: 10175 $
# $Id: //depot/main/qubeMobileServer/qubeMobile/model/qbUIMobileServerUser.py#5 $
# $File: //depot/main/qubeMobileServer/qubeMobile/model/qbUIMobileServerUser.py $
#
# Written by: Brian Knudson / PFX
#             briank@pipelinefx.com
#             2012-06-04
#
# Description:
#  *Very* basic, self-contained user management by way of a single
#  pickled file.  Passwords are not saved as cleartext and are not
#  designed to be unencrypted, but this is not considered a secure
#  password/user manager
#
# Notes:
#
#######################################################################
"""
import os
import sys
import pickle
import base64
import logging
try:
    import qb
except ImportError:
    for api_path in ("/Applications/pfx/qube/api/python/",
                     "/usr/local/pfx/qube/api/python/",
                     "C:\\Program Files (x86)\\pfx\\qube\\api\\python"):
        if api_path not in sys.path and os.path.exists(api_path):
            sys.path.insert(0,api_path)
            break
    import qb


class QubeMobileServerUserDB(object):
    """
    A class to hold insecure username/password data.
    The database will accept cleartext passwords, but it will not
    store them as such.  It will use one-way hashing for password
    comparison
    """
    def __init__(self,password_file=None):
        self.password_file = password_file
        self.__readDb()


    def __del__(self):
        """
        write out the db when closed
        """
        self.__writeDb()

    def __readDb(self):
        """
        [re]reads the database file
        """
        self.__password_database = self.__read_encoded_file(self.password_file)
        if not self.__password_database:
            self.__password_database = {}
            self.__password_database_mtime = 0
        else:
            self.__password_database_mtime = os.path.getmtime(self.password_file)

    def __writeDb(self):
        """
        Writes the database to disk, overwriting the old one
        """
        self.__create_encoded_file(data=self.__password_database,filename=self.password_file)

        
    def __create_encoded_file(self,data,filename):
        """
        Creates an encoded file that can later be decoded.
        The file is obfuscated, but not secure.  Its contents will
        be the base64 encoded representation of the pickeled data.
        """
        try:
            fp = open(filename,'w')
            fp.write(base64.urlsafe_b64encode(pickle.dumps(data)))
            fp.close()
            return True
        except Exception,e:
            logging.warn("Problem writing passwd file '%s': %s" % (filename,e))
            return False
        
    def __read_encoded_file(self,filename):
        """
        Given a filename, this will read and return the file contents.
        It is assumed the file was created with create_encoded_file
        """
        if not os.path.exists(filename):
            return None
        try:
            return pickle.loads(base64.urlsafe_b64decode(open(filename).read()))
        except Exception,e:
            print >>sys.stderr, "Problem reading passwd file '%s': %s" % (filename,e)
            logging.error("Problem reading passwd file '%s': %s" % (filename,e))
            return None

    def __checkDb(self):
        """
        checks mtime for db file. If it's been modified since we last read, re-read.
        """
        if not os.path.exists(self.password_file):
            return 
        if os.path.getmtime(self.password_file) > self.__password_database_mtime:
            self.__readDb()

    @property
    def users(self):
        """
        return a list of user names and encrypted password
        """
        return self.__password_database.items()

    def authenticate(self,user,password):
        """
        returns true if the given user/pass matches our db
        """
        hashed_pw = qb.encryptpassword(password)
        self.__checkDb()
        return hashed_pw == self.__password_database.get(user,'---')

    
    def addUser(self,user,password=None,hashed=False):
        """
        Adds a user to the database & saves the db to disk
        """
        if not hashed:
            password = qb.encryptpassword(password)
            
        self.__password_database[user] = password
        self.__writeDb()

    def dropUser(self,user):
        """
        Adds a user to the database & saves the db to disk
        """
        try:
            logging.info("Dropping user '%s'" % user)
            del(self.__password_database[user])
            self.__writeDb()
        except KeyError:
            pass

    def hasUser(self,user):
        """
        Returns true if the user exists in the db.  Does not check password
        """
        self.__checkDb()
        return self.__password_database.has_key(user)

    def getPassword(self,user):
        """
        returns the hashed password of the given user
        """
        return self.__password_database.get(user)

    def getSize(self):
        """
        Returns the number of entries in the db
        """
        self.__checkDb()
        return len(self.__password_database)
