#!/usr/bin/env python

"""
#######################################################################
#
# Licensed Materials - Property of Pipelinefx L.L.C.
#
#
# (C) COPYRIGHT Pipelinefx Limited Liability Corporation.
#  All Rights Reserved.
#                               
#  US Government Users Restricted Rights - Use, duplication or
#  disclosure restricted by GSA ADP Schedule Contract with 
#  Pipelinefx L.L.C.
#
# $DateTime: 2012/12/17 15:21:39 $
# $Revision: #17 $
# $Change: 10481 $
# $Id: //depot/main/qubeMobileServer/qubeMobileServer.py#17 $
# $File: //depot/main/qubeMobileServer/qubeMobileServer.py $
#
# Written by: Brian Knudson / PFX
#             briank@pipelinefx.com
#             2012-06-04
#
# Description:
#  Python-based web server, designed specifically for qubeMobileView
#
# Notes:
#  - Simply run, no args needed: python webserver.py
#  - Nothing outside of a typical python install is required for normal
#    use, however, qubeMobileView relies on mako templates, so you'll
#    likely need to have python-mako installed.  This should be available
#    via your favorite package manager, or via pip: 'pip install Mako'
#  - Likewise, the models will make use of python MySQLdb if it is
#    available.  Using MySQLdb will allow for less communication
#    and less stress on the supervisor.  It is highly recommended
#    that you instal python MySQLdb.
#
#######################################################################
"""

import cgi
import urlparse
import urllib
import time
import datetime
import os
import sys
import socket
import random
import logging
import tempfile
from optparse import OptionParser, SUPPRESS_HELP
from BaseHTTPServer import BaseHTTPRequestHandler, HTTPServer
from SocketServer import ThreadingMixIn
from Cookie import SimpleCookie

from config import qubeMobileServerVersion
from config.qubeMobileServerSettings import QubeMobileServerSettings
from qubeMobile.model.qbUIMobileServerUser import QubeMobileServerUserDB
from qubeMobile.model.qbUIJobList import QbUIJobList
from qubeMobile.model.qbUIUserList import QbUIUserList
from qubeMobile.controller.qbController import QbController
from util.fileUtils import *

try:
    import pwd
except:
    import win32api

try:
    import mako
except ImportError,e:
    print >>sys.stderr, "Could not import mako templat engine: %s" % e
    print >>sys.stderr, "QubeMobileServer uses mako templates for rendering html pages."
    print >>sys.stderr, "You must install mako templates before you use the server."
    print >>sys.stderr, "If you have pip installed, simple pip install Mako. Otherwise,"
    print >>sys.stderr, "see http://www.makotemplates.org/."
    sys.exit(1)

try:
    import qb
except ImportError:
    for api_path in ("/Applications/pfx/qube/api/python/",
                     "/usr/local/pfx/qube/api/python/",
                     "C:\\Program Files (x86)\\pfx\\qube\\api\\python"):
        if api_path not in sys.path and os.path.exists(api_path):
            sys.path.insert(0,api_path)
            break
    import qb

import warnings 
warnings.filterwarnings("ignore", category=DeprecationWarning)

class QubeMobileServer(ThreadingMixIn, HTTPServer):
    """
    http server that contains a reference to a Qube user
    """
    def __init__(self,
                 use_ssl,
                 sessions_dir,
                 authentication,
                 password_file,
                 allow_control,
                 jobs_in_cache,
                 refresh_interval,
                 template_dir,
                 *args, **kwargs):
        HTTPServer.__init__(self, *args, **kwargs)
        self.sessions = {}
        self.template_dir = template_dir
        self.sessions_dir = sessions_dir
        self.allow_control = allow_control
        self.job_list = QbUIJobList(limit=jobs_in_cache)
        self.refresh_interval = refresh_interval
        self.__authentication = authentication
        self.__password_db = QubeMobileServerUserDB(password_file=password_file)


    def __sessionFromDisk(self,sid):
        """
        Returns the sessions from disk.  If no sessions exists, return None.
        """
        session_file = os.path.join(self.sessions_dir,sid)

        #NOTE: should probably check for readability rather than just existence
        if not os.path.exists(session_file):
            return None

        session = read_encoded_file(session_file)
        self.sessions[sid] = session
        
        return self.sessions[sid]

    @property
    def authentication(self):
        """
        Returns true if we are using password protection
        """
        return self.__authentication

    def hasUser(self,user):
        """
        Returns true if the pw db has this username
        """
        return self.__password_db.hasUser(user)

    def authenticate(self,user,password):
        """
        Tries to authenticate against the password database
        """
        return self.__password_db.authenticate(user,password)
        
        
    def getSession(self,sid):
        """
        Returns the session, given an sid.  If the sessions isn't in
        memory, it attempts to first load the session from disk.
        """
        return self.sessions.get(sid,self.__sessionFromDisk(sid))


    def writeSessions(self):
        """
        Writes all sessions in memory out to disk so they
        can persist across restarts.
        """
        if not os.path.exists(self.sessions_dir):
            try:
                os.makedirs(self.sessions_dir)
            except OSError,e:
                logging.warn("Could not write session files.  Sessions will not persist. %s" % e)
                self.sessions_dir = tempfile.mkdtemp(prefix="qubeMobile-sess_")
                logging.warn("Using temporary storage for sessions: %s." % self.sessions_dir)


        for k,v in self.sessions.items():
            create_encoded_file(data=v,filename=os.path.join(self.sessions_dir,k))


class QubeMobileHandler(BaseHTTPRequestHandler):
    SESSION_COOKIE_NAME = 'qubeMobileServerBiscuit'
    session = None
    controller = None

    def address_string(self):
        host, port = self.client_address[:2]
        #return socket.getfqdn(host)
        return host
  
    def getSession(self):
        """
        Return the existing session or a new session
        """
        if self.session is not None:
            return self.session
  
        # Get value of cookie header that was sent
        cookie_str = self.headers.get('Cookie')
        if cookie_str:
            cookie_obj = SimpleCookie(cookie_str)
            sid_morsel = cookie_obj.get(self.SESSION_COOKIE_NAME, None)
            if sid_morsel is not None:
                sid = sid_morsel.value
            else:
                sid = None
        else:
            sid = None

        # If a session id was not set, create a new one
        if sid is None:
            # maybe overdoing it... maybe not.
            # sid = hashlib.new("sha1",''.join(random.choice('0123456789abcdef') for x in range(16))).hexdigest()
            sid = ''.join(random.choice('0123456789abcdef') for x in range(24))
            session = None
        else:
            session = self.server.getSession(sid)

        self.log_message("Session id: %s", sid)

        # If no session exists for this session ID, create one
        if session is None:
            session = self.server.sessions[sid] = {}
  
        session['id'] = sid
        self.session = session
        return session
  

    def updateSession(self,user,only_my_jobs,logout_requested):
        """
        Update the current session with the given info.
        """
        if not self.session:
            self.getSession()
        self.session['user'] = user
        self.session['only_my_jobs'] = only_my_jobs
        if logout_requested:
            self.session['auth'] = False
        self.server.writeSessions()

    def sendSessionCookie(self):
        if not self.session:
            self.getSession()
        sid = self.session.get('id')
        session_cookie = '%s=%s;' % (self.SESSION_COOKIE_NAME, sid)
        self.send_header('Set-Cookie', session_cookie)
        
    
    def checkAuth(self):
        """
        Returns true if our session has been authorized.
        Authorization is permanent & should persist across server restarts
        """
        # Check to see if we're not using passwords:
        if not self.server.authentication:
            return True

        if not self.session:
            self.getSession()

        return self.session.get("auth",False)

    def checkActionAuth(self,jobid):
        """
        Returns true if the authorized user is eligible to do actions
        This is done be seeing if user and 
        """
        if not self.checkAuth():
            return False

        if self.session.get('auth') != self.session.get('user'):
            return False

        if self.server.job_list.get(jobid,{}).get('user') != self.session.get('auth'):
            return False

        return True

    def checkSessionUser(self):
        """
        Checks if the session user exists in the user db
        """
        user = self.getSessionUser()
        if not user:
            return False
        return self.server.hasUser(user)

    def getSessionUser(self):
        """
        returns the username in this session.  returns None if none.
        """
        if not self.session:
            self.getSession()
        return self.session.get('user')


    def authenticate(self,user,password):
        """
        Checks the password agaist the local password database.
        """
        if not self.session:
            self.getSession()

        self.session['user'] = user
        if self.server.authenticate(user,password):
            self.session['auth'] = user

        self.server.writeSessions()

        
        

    def do_GET(self):
        parsed_uri = urlparse.urlparse(self.path)
        query = {}
        for k, v in cgi.parse_qsl(parsed_uri[4]):
            query[k] = v.decode('utf-8')
        path = parsed_uri[2]
        
        try:
            if path=='/favicon.ico':
                f = open(os.path.join(os.curdir,"qubeMobile","static","favicon.ico"),'rb')
                self.send_response(200)
                self.send_header('Content-type', 'image/x-icon')
                self.end_headers()
                self.wfile.write(f.read())
                f.close()
                return
            
            elif path.endswith(".html"):
                #note that this potentially makes every file on your computer readable by the internet
                f = open(os.curdir + os.sep + path)

                self.send_response(200)
                self.send_header('Content-type', 'text/html')
                self.end_headers()
                self.wfile.write(f.read())
                f.close()
                return
            elif path.endswith(".css"):
                f = open(os.curdir+os.sep+path)
                self.send_response(200)
                self.send_header('Content-type', 'text/css')
                self.end_headers()
                self.wfile.write(f.read())
                f.close()
                return
            elif path.endswith(".js"):
                f = open(os.curdir+os.sep+path)
                self.send_response(200)
                self.send_header('Content-type', 'application/x-javascript')
                self.end_headers()
                self.wfile.write(f.read())
                f.close()
                return
            elif path.endswith(".png"):
                f=open(os.curdir+os.sep+path,'rb')
                self.send_response(200)
                self.send_header('Content-type', 'image/png')
                self.end_headers()
                self.wfile.write(f.read())
                f.close()
                return
            else:   #our dynamic content - may be a bit too 'catch-all' - may want to do some parsing
                if not self.controller:
                    self.controller = QbController(job_list=self.server.job_list,
                                                   refresh_interval=self.server.refresh_interval,
                                                   template_dir=self.server.template_dir)

                self.getSession()

                if self.checkAuth() or path.endswith("authenticate"):
                    self.send_response(200)
                    self.sendSessionCookie()
                    self.send_header('Content-type', 'text/html')
                    self.end_headers()
                    self.wfile.write(self.controller.handle(path=path,session=self.session.copy(),
                                                            allow_control=self.server.allow_control,**query))
                else:  # Send them to the login screen.
                    self.send_response(303)
                    self.send_header("Location","/authenticate")
                    self.end_headers()
                return
                
            return
                
        except IOError:
            self.send_error(404,'File Not Found: %s' % self.path)
     

    def do_POST(self):
        global rootnode
        try:
            parsed_uri = urlparse.urlparse(self.path)
            path = parsed_uri[2]
            ctype, pdict = cgi.parse_header(self.headers.getheader('content-type'))
        
            if ctype == 'application/x-www-form-urlencoded':
                length = int(self.headers.getheader('content-length'))
                postvars = cgi.parse_qs(self.rfile.read(length), keep_blank_values=1)
                # print "postvars:",postvars
                if path == '/authenticate' and postvars.has_key('user') and postvars.has_key('pass'):
                    user = postvars['user'][0]
                    user = user.lower()
                    password = postvars['pass'][0]
                    self.authenticate(user,password)
                    if self.checkAuth():
                        self.send_response(303)
                        self.send_header("Location","/")
                        self.end_headers()
                    else:
                        if self.checkSessionUser():
                            self.send_response(303)
                            self.send_header("Location","/authenticate?reason=%s" % urllib.quote_plus("Password incorrect"))
                            self.end_headers()
                        else:
                            self.send_response(303)
                            self.send_header("Location","/authenticate?reason=%s" % urllib.quote_plus("Unknown user"))
                            self.end_headers()
                elif path == '/settings':
                    user = postvars['user'][0]
                    user = user.lower()
                    only_my_jobs = postvars.get('only_my_jobs',False)
                    logout_requested = postvars.get('logout',False)
                    if logout_requested:
                        logout_requested = True
                    self.updateSession(user,only_my_jobs,logout_requested)

                    self.send_response(303)
                    self.send_header("Location","/settings?reason=%s" % urllib.quote_plus("Settings updated"))
                    self.end_headers()
                    
                elif postvars.get('kill',False):
                    job_id = int(postvars['kill'][0])
                    self.log_message("Request to kill job: %s", job_id)
                    action_message = ""
                    if self.checkActionAuth(job_id):
                        killed = self.server.job_list[job_id].kill()
                        self.log_message("Supervisor response to kill: '%s'" % killed)
                        if killed:
                            action_message = "Killed %s" % job_id
                        else:
                            action_message = "Failed to kill %s" % job_id
                    else:
                        self.log_message("User %s tried to kill %s. Denied." % (self.session.get('auth'),job_id))
                        action_message = "You are not authorized to kill %s" % job_id
                    self.send_response(303)
                    self.send_header("Location","%s%s" % (path,{True:"?msg=%s" % urllib.quote_plus(action_message),
                                                                False:""}[len(action_message) > 1]))
                    self.end_headers()

                elif postvars.get('retry',False):
                    job_id = int(postvars['retry'][0])
                    self.log_message("Request to retry job: %s", job_id)
                    action_message = ""
                    if self.checkActionAuth(job_id):
                        retried = self.server.job_list[job_id].retry(only_failed=True)
                        self.log_message("Supervisor response to retry: '%s'" % retried)
                        if retried:
                            action_message = "Retried failed for %s" % job_id
                        else:
                            action_message = "Failed to retry %s" % job_id
                    else:
                        self.log_message("User %s tried to retry %s. Denied." % (self.session.get('auth'),job_id))
                        action_message = "You are not authorized to retry %s" % job_id

                    self.send_response(303)
                    self.send_header("Location","%s%s" % (path,{True:"?msg=%s" % urllib.quote_plus(action_message),
                                                                False:""}[len(action_message) > 1]))
                    self.end_headers()
                    
                    

            else:
                postvars = {}
                self.log_error("Unhandled POST ctype '%s'.  Ignoring",ctype)
        
            
        except Exception,e:
            self.log_error("An error occured during POST processing: %s",str(e))
            raise


def checkControl():
    """
    Checks to see if the user running this app has admin privileges
    that are necessary to allow users to control their jobs through
    the server
    """
    ul = QbUIUserList()
    username = None

    try:
        username = pwd.getpwuid(os.getuid()).pw_name
    except NameError:
        username = win32api.GetUserName()
    except:
        username = getpass.getuser()

    user_perms = ul.get(username,None)
    if user_perms:
        return user_perms.hasQubePermission('admin')
    return False



def main(settings):
    """
    Web server's main routine.  Running this starts the webserver and will continue
    running it until someone gives a ^c.

    port = the port on which to run the server
    """
    if not settings.get("port"):
        logging.error("No port specified.  Does the conf file exist?")
        sys.exit(1)
    if not settings.get("template_dir"):
        logging.warn("No template dir specified.  The server may not be serving any pages.")
    if not os.path.exists(settings.get("template_dir")):
        logging.warn("Cannot find specified template dir '%s'.  The server may not be serving any pages." % settings.get("template_dir"))

    print "QubeMobileServer v%s (c) PipelineFX, LLC" % qubeMobileServerVersion.VERSION_STRING

    print "Attempting to contact supervisor...",
    sys.stdout.flush()
    supe_info = qb.ping().split()
    if supe_info:
        print " Found supervisor at %s" % supe_info[0]
    else:
        logging.error("Could not contact supervisor.  Is qb.conf set correctly?")
        sys.exit(1)

    try:
        use_ssl = settings.get("use_ssl",False)
        server = QubeMobileServer(use_ssl,
                                  settings.get("sessions_dir"),
                                  settings.get("authentication",False),
                                  settings.get("password_file"),
                                  settings.get("allow_control",False),
                                  settings.get("jobs_in_cache",200),
                                  settings.get("refresh_interval",3600),
                                  settings.get("template_dir"),
                                  ('', settings.get("port")),
                                  QubeMobileHandler)
        ip = ''
        try:
            ip = socket.gethostbyname(socket.gethostname())
        except Exception,e:
            print >>sys.stderr, "Could not determine ip address: %s" % e
            ip = 'UNKNOWN'
        if ip == '127.0.0.1':
            ip = 'UNKNOWN'

        if use_ssl:
            try:
                import ssl
            except ImportError,e:
                print >>sys.stderr, "\nCould not import ssl module needed for running in ssl mode:"
                print >>sys.stderr, e
                print >>sys.stderr, "The ssl module is included with python >= 2.6.  If you have an older"
                print >>sys.stderr, "version of python, you can either add the ssl module, upgrade python,"
                print >>sys.stderr, "or run in non-ssl mode.  Use --no-ssl options to run without ssl.\n"
                sys.exit(1)

            certificate_dir = settings.get("certificate_dir")
            certificate_file = settings.get("certificate_file")
            key_file = settings.get("key_file")

            if not certificate_file or not key_file:
                from util.makeCert import createSelfSignedCert
                if not os.path.exists(certificate_dir):
                    try:
                        os.makedirs(certificate_dir,0700)
                    except OSError:
                        logging.warn("Could not create certificate dirs in %s" %os.path.abspath(certificate_dir))
                        certificate_dir = tempfile.mkdtemp(prefix="qubeMobile-cert_")
                        logging.warn("Using temporary storage for certs: %s." % certificate_dir)

                certificate_file,key_file = createSelfSignedCert(certificate_dir)

            if not os.path.exists(certificate_file) or not os.path.exists(key_file):
                print >>sys.stderr, "Cert files not corrently generated.  Cannot run ssl server."
                sys.exit(1)

            # wrap ssl around our server.  (that was easy)
            server.socket = ssl.wrap_socket (server.socket, keyfile=key_file, certfile=certificate_file, server_side=True)

        # Inform the user what we're about to do
        start_message = 'Starting %squbeMobile httpserver on http%s://%s%s...' % ({True:"*secure* ",False:"-insecure- "}[use_ssl==True],
                                                                                  {True:"s",False:""}[use_ssl==True],
                                                                                  ip,
                                                                                  {True:"",False:":%d"%settings.get("port")}[settings.get("port")==80])
        print >>sys.stderr, start_message

        server.serve_forever()
    except KeyboardInterrupt:
        stop_message = '^C received, shutting down server'
        print >>sys.stderr, stop_message
        logging.info(stop_message)
        logging.info("closing socket")
        server.socket.close()
        logging.info("writing sessions from memory to disk")
        server.writeSessions()
        logging.info("closing server")
        server.server_close()
        logging.info("complete")


def templateDirHelper(template_dir,script_path,launch_path=None):
    """
    Takes the given template dir & attempts to find it as an absolute
    path or a path relative to either the launch or script location(s)
    """
    if not template_dir:
        logging.warning("No template dir specified.  Will try to find one.")
        return False
    if os.path.exists(template_dir):
        return template_dir
    if os.path.exists(os.path.join(script_path,template_dir)):
        return os.path.join(script_path,template_dir)
    if os.path.exists(os.path.join(launch_path,template_dir)):
        return os.path.join(launch_path,template_dir)
    return False
    

if __name__ == '__main__':
    launch_cwd = os.path.abspath(os.getcwd())
    script_location = os.path.abspath(os.path.dirname(sys.argv[0]))
    if script_location != launch_cwd:
        os.chdir(script_location)

    settings = QubeMobileServerSettings(os.path.join(script_location,"config","qbms.conf"))

    default_settings_file = os.path.abspath(settings.conf_file),
    default_port =          settings.get("port",8080)
    default_certificate_dir =      settings.get("certificate_dir",os.path.join(launch_cwd,"certs"))
    default_certificate_file =     settings.get("certificate_file")
    default_key_file =      settings.get("key_file")
    default_sessions_dir =  settings.get("sessions_dir",os.path.join(launch_cwd,"sessions"))
    default_password_file = settings.get("password_file",os.path.join(launch_cwd,".passwd"))
    default_ssl =           settings.get("use_ssl",False)
    default_authentication= settings.get("authentication")
    default_allow_control = settings.get("allow_control",False)
    default_jobs_in_cache = settings.get("jobs_in_cache",200)
    default_refresh_interval = settings.get("refresh_interval",3600)
    default_template_dir  = settings.get("template_dir")

    certificate_help = "Use this directory for storing self-signed certificates needed for ssl.  "
    certificate_help += "If the cert doesn't exist, it will be created for you. Default: '%s'.  " % default_certificate_dir
    certificate_help += "*Note: pyOpenSSL required."

    session_help  = "Directory to store session information so sessions can persist "
    session_help += "across restarts.  Default: '%s'" % default_sessions_dir

    passwd_help  = "Location of the server-created password file.  "
    passwd_help += "Note this should absolutely NOT be a system-managed password file. "
    passwd_help += "Default: '%s'." % default_password_file
    
    try:
        version_string = qubeMobileServerVersion.VERSION_STRING
    except:
        version_string = 'unknown'
    p = OptionParser(version=version_string)
    p.add_option("",  "--settings-file",
                 dest="settings_file",
                 default=default_settings_file,
                 action="store",
                 type="string",
                 help="Location of the qbms.conf file. Default: %s" % default_settings_file)
    p.add_option("",  "--ssl",
                 dest="use_ssl",
                 default=default_ssl,
                 action="store_const",
                 const=True,
                 help="Use SSL for communication. Default OS X/Linux: use SSL; "
                 "Windows will ignore this option. *Note: ssl module required. "
                 "This modules ships with python 2.6 and newer.")
    if sys.platform == 'win32':
        force_ssl_help = "On Windows, force the use of SSL. Note, this requires "\
                         "that you have your own certificates and may cause the "\
                         "server to stop serving pages."
    else:
        force_ssl_help = SUPPRESS_HELP
    p.add_option("",  "--force-ssl",
                 dest="force_ssl",
                 default=False,
                 action="store_true",
                 help=force_ssl_help)
    p.add_option("",  "--no-ssl",
                 dest="use_ssl",
                 default=default_ssl,
                 action="store_const",
                 const=False,
                 help="Do not use ssl. Default: use SSL. *Note: ssl module required for SSL.")
    p.add_option("-p",  "--port",
                 dest="port",
                 type="int",
                 default=default_port,
                 help="Port on which to run. Default: %d" % default_port)
    p.add_option("-j",  "--jobs-in-cache",
                 dest="jobs_in_cache",
                 type="int",
                 default=default_jobs_in_cache,
                 help="Number of jobs in server cache. Default: %d" % default_jobs_in_cache)
    p.add_option("-i",  "--refresh-interval",
                 dest="refresh_interval",
                 type="int",
                 default=default_refresh_interval,
                 help="Amount of time after refresh before cache is considered stale. "
                 "Default: %d" % default_refresh_interval)
    p.add_option("-d",  "--certificate-dir",
                 dest="certificate_dir",
                 default=default_certificate_dir,
                 action="store",
                 type="string",
                 help=certificate_help)
    p.add_option("-c", "--certificate-file",
                 dest="certificate_file",
                 default=default_certificate_file,
                 action="store",
                 type="string",
                 help="Path to PEM cert file if you have your own.")
    p.add_option("-k", "--key-file",
                 dest="key_file",
                 default=default_key_file,
                 action="store",
                 type="string",
                 help="Path to key file if you have your own.")
    p.add_option("-s", "--session-dir",
                 dest="sessions_dir",
                 default=default_sessions_dir,
                 action="store",
                 type="string",
                 help=session_help)
    p.add_option("", "--authentication",
                 dest="authentication",
                 default=default_authentication,
                 action="store_const",
                 const=True,
                 help="Require users to authenticate against MobileServer's "
                 "password database.  Default: no authentication")
    p.add_option("", "--no-authentication",
                 dest="authentication",
                 default=default_authentication,
                 const=False,
                 action="store_const",
                 help="Do not require users to supply passwords.  A user can log "
                 "on with username only.  Default: no authentication")
    p.add_option("", "--password-file",
                 dest="password_file",
                 default=default_password_file,
                 action="store",
                 type="string",
                 help=passwd_help)
    p.add_option("", "--template-dir",
                 dest="template_dir",
                 default=default_template_dir,
                 action="store",
                 type="string",
                 help="Path to the html template directory.")
    p.add_option("", "--allow-control",
                 dest="allow_control",
                 default=default_allow_control,
                 action="store_true",
                 help="Allow users to control their jobs. Note: This server "
                 "must be run by a Qube administrator to use this option. Setting "
                 "this options implies authentication and ssl.")

    options,args = p.parse_args(sys.argv)

    if options.use_ssl and sys.platform == "win32":
        if options.force_ssl:
            pass
        else:
            print >>sys.stderr, "\nWARNING: You cannot run over ssl in Windows.  The server will run *without* ssl.\n"
            options.use_ssl=False

    if options.force_ssl:
        options.use_ssl = True

    if options.settings_file != default_settings_file:
        settings = QubeMobileServerSettings(options.settings_file)

    if options.certificate_dir != default_certificate_dir and (options.certificate_file or options.key_file):
        message  = "\nYou cannot specify a certificate directory AND certificate file or key.\n"
        message += "If you already have a certificate file/key, specify only those.\n"
        message += "If you do not have a certificate file/key, specify a cert directory and this\n"
        message += " app will generate a self-signed certificate/key for you.\n"
        p.error(message)

    # If allow control requiested, assume the user knows what they are doing & force other options.
    if options.allow_control:
        if not options.use_ssl:
            print >>sys.stderr, " - Forcing ssl to allow control -"
            options.use_ssl = True
        if not options.authentication:
            print >>sys.stderr, " - Forcing authentication to allow control -"
            options.authentication = True

    # If a password file is given, assume the user knows what they are doing & force other options.
    if options.password_file != default_password_file:
        if not options.authentication:
            print >>sys.stderr, " - Forcing authentication because password file was specified -"
            options.authentication = True
        if not options.use_ssl:
            print >>sys.stderr, " - Forcing ssl because password file was specified -"
            options.ssl = True

    if options.allow_control:
        if not checkControl():
            print >>sys.stderr, "This user does not have Qube administrative privileges."
            print >>sys.stderr, "You cannot allow user control unless this server is run by a Qube! admin."
            sys.exit(1)

    # update settings from user options
    settings.update(vars(options))

    template_dir = templateDirHelper(settings.get("template_dir"),script_location,launch_cwd)
    if template_dir:
        settings.template_dir = template_dir

    main(settings=settings)

